<?php

/*!
 * C-Rent Web
 *
 * Copyright 2013, CX9 Systems GmbH
 *
 * Authors: Christoph Weitkamp - c.weitkamp@cx9.de
 * Authors: Yuriy Molodyukov - y.molodyukov@cx9.de
 *
 * http://www.cx9.de
 */

require_once "DBConnection.php";

if (!defined('DB_DEBUG')) {
    define('DB_DEBUG', false);
}

if (!defined('SCRIPT_DIE')) {
    define('SCRIPT_DIE', true);
}

if (!defined('TRANSACT_LEERLAUF')) {
    define('TRANSACT_LEERLAUF', false);
}

if (!defined('PERSISTENT')) {
    define('PERSISTENT', false);
}

define("OS_LINUX", 0);
define("OS_WINDOWS", 1);
define('SYSTEM_OS', strtoupper(substr(PHP_OS, 0, 3)) === 'WIN' ? OS_WINDOWS : OS_LINUX);

class DBPDOConnect extends DBConnection
{
    //
    private $dbh = null;
    private $sth = null;
    private $sqlDebug = '';
    private $result_stack = array();

    public function __construct($host, $username, $password, $dbname, $instance)
    {
        // call constructor of parent class
        parent::__construct($host, $username, $password, $dbname, $instance);
    }

    public function __sleep()
    {
        return array('host', 'username', 'password', 'dbname');
    }

    public function __wakeup()
    {
        //parent::__construct($this->host, $this->username, $this->password, $this->dbname);
    }

    /**
     * Stops the execution of the script, if constant SCRIPT_DIE = true
     *
     */
    private function script_die()
    {
        if (SCRIPT_DIE) {
            die();
        }
    }

    /**
     * @param persistent bool Persistent connections
     * @return true|false
     */
    public function connect($persistent = false)
    {
        $connect_flag = false;
        // disconnect from db
        $this->disconnect();
        // attributes for db connection
        $attr = array(PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION);
        if ($persistent) {
            $attr[PDO::ATTR_PERSISTENT] = true;
        }

        $pdo_driver_found = false;
        try {
            $this->dbh = new PDO(((SYSTEM_OS === OS_LINUX) ? "dblib" : "mssql") . ":host=$this->host;dbname=$this->dbname", $this->username, $this->password);
            $pdo_driver_found = true;
            $connect_flag = true;
        } catch (PDOException $e) {
            if (DB_DEBUG) {
                echo $e->getMessage();
            } else {
                echo "Die Datenbank ist aktuell nicht erreichbar (#A)";
            }
            $this->log("DBPDOConnect:connect()", $e->getMessage());
            $this->dbh = null;

            $this->script_die();
        }

        if (!$pdo_driver_found) {
            try {
                $this->dbh = new PDO("sqlsrv:Server=$this->host;Database=$this->dbname", $this->username, $this->password);
                $pdo_driver_found = true;
                $connect_flag = true;
            } catch (PDOException $e) {
                if (DB_DEBUG) {
                    echo $e->getMessage();
                } else {
                    echo "Die Datenbank ist aktuell nicht erreichbar (#B)";
                }
                $this->log("DBPDOConnect:connect()", $e->getMessage());
                $this->dbh = null;

                $this->script_die();
            }
        }

        return $connect_flag;
    }

    /**
     *  Disconnect current connection
     */
    public function disconnect()
    {
        if (isset($this->dbh)) {
            unset($this->dbh);
        }
    }

    /**
     * Executes an SQL statement, returning a result set as a PDOStatement object
     * @param sql string <p>
     * The SQL statement to prepare and execute.
     * </p>
     * <p>
     * Data inside the query should be properly escaped.
     * </p>
     * @return PDOStatement PDO::query returns a PDOStatement object, or false
     * on failure.
     */
    public function query($sql)
    {
        try {
            if (!isset($this->dbh)) {
                throw new PDOException("Keine Verbindung zur Datenbank!");
            }
            if (trim(strlen($sql)) > 0) {
                // set language to german
                @$this->dbh->exec("SET LANGUAGE deutsch");

                //
                $this->sth = $this->dbh->query($sql);
                if (!$this->sth) {
                    throw new PDOException("Fehler in der Abfrage!\n" . $sql . "\n" . print_r($this->dbh->errorInfo(), true));
                }
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            $this->log("DBPDOConnect:query()", $e->getMessage());

            if (isset($this->dbh) && $this->dbh->inTransaction()) {
                throw $e;
            } else {
                if (DB_DEBUG) {
                    echo "Error: DBPDOConnect:query() -> " . $e->getMessage();
                } else {
                    echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
                }

                $this->script_die();
                $this->sth = null;
            }
        }
        return $this->sth;
    }

    /**
     * Execute an SQL query and return the number of affected rows
     * @param sql string <p>
     * </p>
     * <p>
     * Data inside the query should be properly escaped.
     * </p>
     * @return int returns the number of rows that were modified
     * or deleted by the SQL query you issued. If no rows were affected,
     * returns 0.
     * </p>
     */
    public function exec($sql)
    {
        try {
            if (!isset($this->dbh)) {
                throw new PDOException("Keine Verbindung zur Datenbank!");
            }
            if (trim(strlen($sql)) > 0) {
                $result = false;
                // set language to german
                @$this->dbh->exec("SET LANGUAGE deutsch");

                //
                $result = $this->dbh->exec($sql);
                if ($result === false) {
                    throw new PDOException("Fehler in der Abfrage!\n" . $sql . "\n" . print_r($this->dbh->errorInfo(), true));
                }
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            $this->log("DBPDOConnect:exec()", $e->getMessage());

            if (isset($this->dbh) && $this->dbh->inTransaction()) {
                throw $e;
            } else {
                if (DB_DEBUG) {
                    echo "Error: DBPDOConnect:exec() -> " . $e->getMessage();
                } else {
                    echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
                }
                $this->script_die();
            }
        }
        return $result;
    }

    /**
     * Executes a prepared statement
     * @return bool Returns true on success or false on failure.
     */

    public function execute($params = array())
    {
        try {
            if (isset($this->sth)) {
                // set language to german
                @$this->dbh->exec("SET LANGUAGE deutsch");

                if (is_array($params) && count($params) > 0) {
                    $result = $this->sth->execute($params);
                } else {
                    $result = $this->sth->execute();
                }

                if (!$result) {
                    throw new PDOException("Fehler in der Abfrage!\n" . $this->sth->queryString . "\n" . print_r($this->sth->errorInfo(), true));
                }
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            if (DEBUG) {
                echo '<pre>';
                print_r($e);
                echo '</pre>';
            }
            $this->log("DBPDOConnect:execute()", $e->getMessage());
            if (isset($this->dbh) && $this->dbh->inTransaction()) {
                throw $e;
            } else {
                if (DEBUG) {
                    echo '<pre>';
                    print_r($this->sth->queryString);
                    echo '</pre>';
                    // var_dump(debug_backtrace()[0]);
                    echo "<br/>Error: DBPDOConnect:execute() -> " . $e->getMessage() . ': ' . $this->sqlDebug . '<br/>';
                } else {
                    if (DEBUG) {
                        echo '<pre>';
                        print_r($this->sth->queryString);
                        echo '</pre>';
                    }
                    echo 'Error: Bei der Abfrage ist ein Fehler aufgetreten!';
                }
                $result = false;
                $this->script_die();
            }
        }

        return $result;
    }

    /**
     * Prepares a statement for execution and returns a statement object
     * @param statement string <p>
     * This must be a valid SQL statement for the target database server.
     * </p>
     * @return PDOStatement If the database server successfully prepares the statement,
     * PDO::prepare returns a
     * PDOStatement object.
     * If the database server cannot successfully prepare the statement,
     * PDO::prepare returns false or emits
     * PDOException (depending on error handling).
     * </p>
     */
    public function prepare($sql)
    {
        try {
            if (!isset($this->dbh)) {
                throw new PDOException("Keine Verbindung zur Datenbank!");
            }
            if (trim(strlen($sql)) > 0) {
                //
                $this->sth = @$this->dbh->prepare($sql);
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            $this->log("DBPDOConnect:prepare()", $e->getMessage());
            if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            } else {
                if (DB_DEBUG) {
                    echo "Error: DBPDOConnect:prepare() -> " . $e->getMessage();
                } else {
                    echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
                }
            }
            $this->sth = null;
            $this->script_die();
        }
        return $this->sth;
    }

    /**
     * Binds a parameter to the specified variable name
     * @param parameter mixed <p>
     * Parameter identifier. For a prepared statement using named
     * placeholders, this will be a parameter name of the form
     * :name. For a prepared statement using
     * question mark placeholders, this will be the 1-indexed position of
     * the parameter.
     * </p>
     * @param variable mixed <p>
     * Name of the PHP variable to bind to the SQL statement parameter.
     * </p>
     * @param data_type int[optional] <p>
     * Explicit data type for the parameter using the PDO::PARAM_*
     * constants.
     * </p>
     * @return bool Returns true on success or false on failure.
     */
    public function bindParam($parameter, &$variable, $data_type = PDO::PARAM_STR)
    {
        try {
            if (isset($this->sth)) {
                $result = @$this->sth->bindParam($parameter, $variable, $data_type);
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            }

            if (DB_DEBUG) {
                echo "Error: DBPDOConnect:bindParam() -> " . $e->getMessage();
            } else {
                echo "Error: Beim Binden eines Parameters ist ein Fehler aufgetreten!";
            }

            $this->log("DBPDOConnect:bindParam()", $e->getMessage());
            $result = false;
            $this->script_die();
        }
        return $result;
    }

    /**
     * Binds a value to a parameter
     * @param parameter mixed <p>
     * Parameter identifier. For a prepared statement using named
     * placeholders, this will be a parameter name of the form
     * :name. For a prepared statement using
     * question mark placeholders, this will be the 1-indexed position of
     * the parameter.
     * </p>
     * @param value mixed <p>
     * The value to bind to the parameter.
     * </p>
     * @param data_type int[optional] <p>
     * Explicit data type for the parameter using the PDO::PARAM_*
     * constants.
     * </p>
     * @return bool Returns true on success or false on failure.
     */
    public function bindValue($parameter, $value, $data_type = PDO::PARAM_STR)
    {
        try {
            if (isset($this->sth)) {
                $result = @$this->sth->bindValue($parameter, $value, $data_type);
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            }

            if (DB_DEBUG) {
                echo "Error: DBPDOConnect:bindValue() -> " . $e->getMessage();
            } else {
                echo "Error: Beim Binden eines Parameters ist ein Fehler aufgetreten!";
            }

            $this->log("DBPDOConnect:bindValue()", $e->getMessage());
            $result = false;
            $this->script_die();
        }
        return $result;
    }

// TODO Sauber machen
    public function bindColumn($col, &$param)
    {
        //$result = @$this->sth->bindColumn($col, $param, PDO::PARAM_LOB, 0, PDO::SQLSRV_ENCODING_BINARY);
        $result = @$this->sth->bindColumn($col, $param, PDO::PARAM_LOB);
    }

    /**
     * returns an array indexed by column number as returned in your result set, starting at column 0
     */
    public function fetch_row()
    {
        if (isset($this->sth)) {
            return $this->sth->fetch(PDO::FETCH_NUM);
        }
        return false;
    }

    public function fetch_bound()
    {
        if (isset($this->sth)) {
            return $this->sth->fetch(PDO::FETCH_BOUND);
        }
        return false;
    }

    /**
     * Returns an array indexed by column number and by column name as returned in your result set, starting at column 0
     */
    public function fetch_array()
    {
        if (isset($this->sth)) {
            return $this->sth->fetch(PDO::FETCH_BOTH);
        }
        return false;
    }

    /**
     * returns an array indexed by column name as returned in your result set
     */
    public function fetch_assoc()
    {
        if (isset($this->sth)) {
            return $this->sth->fetch(PDO::FETCH_ASSOC);
        }
        return false;
    }

    /**
     * Returns an array containing all of the result set rows
     * @param param string Fetch style
     * <p>
     *     'assoc' (default) returns an array indexed by column name as returned in your result set
     * </p>
     * <p>
     *     'both' returns an array indexed by both column name and 0-indexed column number as returned in your result set
     * </p>
     * <p>
     *     'num'  returns an array indexed by column number as returned in your result set, starting at column 0
     * </p>
     * <p>
     *     'obj'  returns an anonymous object with property names that correspond to the column names returned in your result set
     * </p>
     */
    public function fetch_all($param = 'assoc')
    {
        if (isset($this->sth)) {
            try {
                switch ($param) {
                    case 'assoc':
                        return $this->sth->fetchAll(PDO::FETCH_ASSOC);
                        break;

                    case 'num':
                        return $this->sth->fetchAll(PDO::FETCH_NUM);
                        break;

                    case 'both':
                        return $this->sth->fetchAll(PDO::FETCH_BOTH);
                        break;

                    case 'obj':
                        return $this->sth->fetchAll(PDO::FETCH_OBJ);
                        break;

                    default:
                        throw new PDOException("Falshe Parametr '" . $param . "'");
                }
            } catch (PDOException $e) {
                if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                    throw $e;
                }

                if (DB_DEBUG) {
                    echo "Error: DBPDOConnect:fetch_all() -> " . $e->getMessage();
                } else {
                    echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
                }

                $this->log("DBPDOConnect:fetch_all()", $e->getMessage());
                $this->script_die();
            }
        }
        return false;
    }

    /**
     * Returns an anonymous object with property names that correspond to the column names returned in your result set
     */
    public function fetch_obj()
    {
        if (isset($this->sth)) {
            return $this->sth->fetch(PDO::FETCH_OBJ);
        }
        return false;
    }

    /**
     * Fetches the next row and returns it as an object.
     * @param class string<p>
     * Name of the created class.
     * </p>
     * @return mixed an instance of the required class with property names that
     * correspond to the column names.
     */
    public function fetch_class($class)
    {
        if (isset($this->sth)) {
            try {
                if (!class_exists($class)) {
                    throw new PDOException('Class "' . $class . '" existirt nicht!');
                } else {
                    $this->sth->setFetchMode(PDO::FETCH_CLASS, $class);
                    $sth_tmp = $this->sth->fetch();
                    $this->sth->setFetchMode(PDO::ATTR_DEFAULT_FETCH_MODE);
                    return $sth_tmp;
                }
            } catch (PDOException $e) {
                if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                    throw $e;
                }

                if (DB_DEBUG) {
                    echo "Error: DBPDOConnect:fetch_class() -> " . $e->getMessage();
                } else {
                    echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
                }

                $this->log("DBPDOConnect:fetch_class()", $e->getMessage());
            }
        }
        return false;
    }

    /**
     * stores the current result in the stack
     */
    public function result_push()
    {
        array_push($this->result_stack, $this->sth);
    }

    /**
     * takes out the current result from the  stack
     */
    public function result_pop()
    {
        if (count($this->result_stack)) {
            $this->sth = array_pop($this->result_stack);
        }
    }

    /**
     * Returns the ID of the last inserted row or sequence value <br>
     * in case of failure returns False
     */
    public function last_insert_id()
    {
        if (isset($this->dbh)) {
            return $this->dbh->lastInsertId();
        }

        return false;
    }

    /**
     * Returns the number of rows affected by the last SQL statement
     * @return int the number of rows.
     */
    public function row_count()
    {
        if (isset($this->sth)) {
            return $this->sth->rowCount();
        }
        return false;
    }

    /**
     * Quotes a string for use in a query.
     * @param str string <p>
     * The string to be quoted.
     * </p>
     * @param parameter_type int[optional] <p>
     * Provides a data type hint for drivers that have alternate quoting styles.
     * </p>
     * @return string a quoted string that is theoretically safe to pass into an
     * SQL statement. Returns false if the driver does not support quoting in
     * this way.
     */
    public function quote($str)
    {
        try {
            if (!isset($this->dbh)) {
                throw new PDOException("Keine Verbindung zur Datenbank!");
            }

            return substr($this->dbh->quote($str), 1, strlen($this->dbh->quote($str)) - 2);
        } catch (PDOException $e) {
            if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            }

            if (DB_DEBUG) {
                echo "Error: DBPDOConnect:real_escape_string() -> " . $e->getMessage();
            } else {
                echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
            }
            $this->log("DBPDOConnect:real_escape_string()", $e->getMessage());
            $this->script_die();
        }
        return false;
    }

    /**
     * Executes stored procedure
     * @param proc_name Name of stored procedure
     * @param params Array with couple $param => $value of params of procedure
     */
    public function proc_exec($proc_name, $params = array())
    {
        try {
            if (isset($this->dbh)) {
                // set language to german
                @$this->dbh->exec("SET LANGUAGE deutsch");

                $sql = 'EXEC ' . $proc_name . ' ';
                $sqlDebug = 'EXEC ' . $proc_name . ' ';
                foreach ($params as $key => $value) {
                    $sql .= ($key['0'] === '@' ? '' : '@') . $key . ' = :' . $key . ', ';
                    $sqlDebug .= '@' . $key . ' = ' . "'" . $value . "', ";
                }
                $this->sqlDebug = $sqlDebug;
                $sql = rtrim($sql, ', ');

                $this->sth = $this->dbh->prepare($sql);

                foreach ($params as $key => $value) {
                    $bool = $this->sth->bindValue($key, $value);
                }

                $result = $this->sth->execute();

                if ($result === false) {
                    throw new PDOException("Fehler in der Abfrage!\n" . $sql . "\n" . print_r($this->sth->errorInfo(), true));
                }
            } else {
                throw new PDOException("Die Abfrage ist leer!");
            }
        } catch (PDOException $e) {
            $this->log("DBPDOConnect:proc_exec()", $e->getMessage());
            if (isset($this->dbh) && $this->dbh->inTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            } else {
                if (DB_DEBUG) {
                    echo "Error: DBPDOConnect:proc_exec() -> " . $e->getMessage();
                    trace();
                } else {
                    echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
                }
            }
            $this->script_die();
        }

        return $result;
    }

    /**
     * Logs the error message with the current time
     * @param function Name of function
     * @param msg Error message
     */
    private function log($function, $msg)
    {
        if (strlen($msg) > 0) {
            file_put_contents("DBPDOConnect.log", date("d-m-Y H:i:s") . ": (" . $function . ") " . $msg . "\n", FILE_APPEND);
        }
    }

    /**
     * Dump a SQL prepared command
     * @return bool
     */
    public function debug_dump_params()
    {
        if (isset($this->sth)) {
            return $this->sth->debugDumpParams();
        }
        return false;
    }

    /**
     * Return an array of available PDO drivers
     * @return array PDO::getAvailableDrivers returns an array of PDO driver names. If
     * no drivers are available, it returns an empty array.
     */
    public static function get_available_drivers()
    {
        return PDO::getAvailableDrivers();
    }

    /** !!! DEPRECATED !!!
     * Always true, because database selection takes place in the constructor
     */
    public function selectDatabase()
    {
        return true;
    }

    /**
     * Fetch extended error information associated with the last operation on the statement handle
     * @return arrays PDO::errorInfo and PDOStatement::errorInfo returns an array of
     * error information about the last operation performed by this
     * statement handle. The array consists of the following fields:
     * <tr valign="top">
     * <td>Element</td>
     * <td>Information</td>
     * </tr>
     * <tr valign="top">
     * <td>0</td>
     * <td>SQLSTATE error code (a five characters alphanumeric identifier defined
     * in the ANSI SQL standard).</td>
     * </tr>
     * <tr valign="top">
     * <td>1</td>
     * <td>Driver specific error code.</td>
     * </tr>
     * <tr valign="top">
     * <td>2</td>
     * <td>Driver specific error message.</td>
     * </tr>
     */
    public function last_message()
    {
        try {
            if (!isset($this->dbh) && !isset($this->sth)) {
                throw new PDOException("Databank Object existiert nicht!");
            }

            ob_start();
            echo "DBH last message: ";
            print_r($this->dbh->errorInfo());

            if (isset($this->sth)) {
                echo "STH last message: ";
                print_r($this->sth->errorInfo());
            }

            $result = "<pre>" . ob_get_contents() . "</pre>";
            ob_end_clean();

            return $result;
        } catch (PDOException $e) {
            if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            }

            if (DB_DEBUG) {
                echo "Error: DBPDOConnect:last_message() -> " . $e->getMessage();
            } else {
                echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
            }
            $this->log("DBPDOConnect:proc_exec()", $e->getMessage());
            $this->script_die();
        }
    }

    public function beginTransaction()
    {
        $this->dbh->beginTransaction();
    }

    public function commit()
    {
        $this->dbh->commit();
    }

    public function rollBack()
    {
        $this->dbh->rollBack();
    }

    /**
     * !!! DEPRECATED !!! <br>
     * Quotes a string for use in a query.
     * @param str string <p>
     * The string to be quoted.
     * </p>
     * @param parameter_type int[optional] <p>
     * Provides a data type hint for drivers that have alternate quoting styles.
     * </p>
     * @return string a quoted string that is theoretically safe to pass into an
     * SQL statement. Returns false if the driver does not support quoting in
     * this way.
     */
    public function real_escape_string($str)
    {
        try {
            if (!isset($this->dbh)) {
                throw new PDOException("Keine Verbindung zur Datenbank!");
            }

            return substr($this->dbh->quote($str), 1, strlen($this->dbh->quote($str)) - 2);
        } catch (PDOException $e) {
            if (isset($this->dbh) && $this->dbh->InTransaction() && TRANSACT_LEERLAUF) {
                throw $e;
            }

            if (DB_DEBUG) {
                echo "Error: DBPDOConnect:real_escape_string() -> " . $e->getMessage();
            } else {
                echo "Error: Bei der Abfrage ist ein Fehler aufgetreten!";
            }
            $this->log("DBPDOConnect:real_escape_string()", $e->getMessage());
            $this->script_die();
        }
        return false;
    }

    /**
     * !!! DEPRECATED !!!
     */
    public function stack_query($sql)
    {
        array_push($this->result_stack, $this->sth);
        $this->sth = $this->dbh->query($sql);
        return $this->sth;
    }

    /**
     * !!! DEPRECATED !!!
     *
     */
    public function stack_free_result()
    {
        //
        $this->free_result();
        if (count($this->result_stack) > 0) {
            $this->sth = array_pop($this->result_stack);
        }
    }

    /**
     * !!! DEPRECATED !!!
     *
     */
    public function free_result()
    {
        if (isset($this->sth)) {
            @$this->sth->closeCursor();
        }
        unset($this->sth);
    }
} // End of Class DBPDOConnect
